"""Lexer functions"""
import string
from collections import deque
from typing import Deque


class TokenizeError(Exception):
    """Error thrown when an invalid JSON string is tokenized"""


def extract_string(json_string: str, index: int, tokens: Deque[str]) -> int:
    """Extracts a single string token from JSON string"""
    start = index
    end = len(json_string)
    index += 1

    while index < end:
        char = json_string[index]

        if char == '\\':
            if index + 1 == end:
                raise TokenizeError("Incomplete escape at end of string")

            index += 2
            continue

        if char == '"':
            index += 1
            string_token = json_string[start:index]
            tokens.append(string_token)
            return index

        index += 1

    raise TokenizeError("Expected end of string")


def extract_number(json_string: str, index: int, tokens: Deque[str]) -> int:
    """Extracts a single number token (eg. 42, -12.3) from JSON string"""
    start = index
    end = len(json_string)

    leading_minus_found = False
    decimal_point_found = False

    while index < end:
        char = json_string[index]
        if char == '.':
            if decimal_point_found:
                raise TokenizeError("Too many decimal points in number")

            decimal_point_found = True

        elif char == '-':
            if leading_minus_found:
                raise TokenizeError("Minus sign in between number")

            leading_minus_found = True

        elif not char.isdigit():
            string_token = json_string[start:index]
            tokens.append(string_token)
            return index

        index += 1

    return index


def extract_special(json_string: str, index: int, tokens: Deque[str]) -> int:
    """Extracts true, false and null from JSON string"""
    end = len(json_string)

    word = ''
    while index < end:
        char = json_string[index]
        if not char.isalpha():
            break

        word += char
        index += 1

    if word in ('true', 'false', 'null'):
        tokens.append(word)
        return index

    raise TokenizeError(f"Unknown token found: {word}")


def tokenize(json_string: str) -> Deque[str]:
    """Converts a JSON string into a queue of tokens"""
    tokens: Deque[str] = deque()

    index = 0
    end = len(json_string)
    while index < end:
        char = json_string[index]

        if char in string.whitespace:
            index += 1

        elif char in '[]{}:,':
            tokens.append(char)
            index += 1

        elif char == '"':
            index = extract_string(json_string, index, tokens)

        elif char == '-' or char.isdigit():
            index = extract_number(json_string, index, tokens)

        else:
            index = extract_special(json_string, index, tokens)

    return tokens



"""Parser functions"""
from ast import literal_eval
from typing import Deque, Dict, List, Union

from json_parser.lexer import tokenize

JSONArray = List[object]
JSONObject = Dict[str, object]
JSONNumber = Union[int, float]


class ParseError(Exception):
    """Error thrown when an invalid JSON tokens is parsed"""


def parse_object(tokens: Deque[str]) -> JSONObject:
    """Parses an object out of JSON tokens"""
    obj: JSONObject = {}

    # special case:
    if tokens[0] == '}':
        tokens.popleft()
        return obj

    while tokens:
        token = tokens.popleft()

        # least amount of tokens left should be a colon, a token and a }
        if len(tokens) < 3:
            raise ParseError("Unexpected end of file while parsing")

        if not token.startswith('"'):
            raise ParseError(f"Expected string key for object, found {token}")

        key = parse_string(token)

        token = tokens.popleft()
        if token != ':':
            raise ParseError(f"Expected colon, found {token}")

        value = _parse(tokens)
        obj[key] = value

        if not tokens:
            raise ParseError("Unexpected end of file while parsing")

        token = tokens.popleft()
        if token not in ',}':
            raise ParseError(f"Expected ',' or '}}', found {token}")

        if token == '}':
            break

    return obj


def parse_array(tokens: Deque[str]) -> JSONArray:
    """Parses an array out of JSON tokens"""
    array: JSONArray = []

    # special case:
    if tokens[0] == ']':
        tokens.popleft()
        return array

    while tokens:
        # least number of tokens left should be a token and a ]
        if len(tokens) < 2:
            raise ParseError("Unexpected end of file while parsing")

        value = _parse(tokens)
        array.append(value)

        token = tokens.popleft()
        if token not in ',]':
            raise ParseError(f"Expected ',' or ']', found {token}")

        if token == ']':
            break

    return array


def parse_string(token: str) -> str:
    """Parses a string out of a JSON token"""
    chars: List[str] = []

    index = 1
    end = len(token) - 1
    while index < end:
        char = token[index]

        if char != '\\':
            chars.append(char)
            index += 1
            continue

        next_char = token[index+1]
        if next_char == 'u':
            hex_string = token[index+2:index+6]
            try:
                unicode_char = literal_eval(f'"\\u{hex_string}"')
            except ValueError as err:
                raise ParseError(f"Invalid escape: \\u{hex_string}") from err

            chars.append(unicode_char)
            index += 6
            continue

        if next_char in '"\\/':
            chars.append(next_char)
        elif next_char == 'b':
            chars.append('\b')
        elif next_char == 'f':
            chars.append('\f')
        elif next_char == 'n':
            chars.append('\n')
        elif next_char == 'r':
            chars.append('\r')
        elif next_char == 't':
            chars.append('\t')
        else:
            raise ParseError(f"Unknown escape sequence found in {token}")

        index += 2

    string = ''.join(chars)
    return string


def parse_number(token: str) -> JSONNumber:
    """Parses a number out of a JSON token"""
    try:
        if token.isdigit():
            number: JSONNumber = int(token)
        else:
            number = float(token)
        return number

    except ValueError as err:
        raise ParseError(f"Invalid token: {token}") from err


def _parse(tokens: Deque[str]) -> object:
    """Recursive JSON parse implementation"""
    token = tokens.popleft()

    if token == '[':
        return parse_array(tokens)

    if token == '{':
        return parse_object(tokens)

    if token.startswith('"'):
        return parse_string(token)

    if token[0] == '-' or token[0].isdigit():
        return parse_number(token)

    special_tokens = {
        'true': True,
        'false': False,
        'null': None,
    }
    if token in special_tokens:
        return special_tokens[token]

    raise ParseError(f"Unexpected token: {token}")


def parse(json_string: str) -> object:
    """Parses a JSON string into a Python object"""
    tokens = tokenize(json_string)

    value = _parse(tokens)
    if len(tokens) != 0:
        raise ParseError(f"Invalid JSON at {tokens[0]}")

    return value



"""JSON parser tests"""
from typing import Dict
import pytest

import json_parser


@pytest.mark.parametrize(
    ('json_string', 'expected'),
    (
        ('{}', {}),
        ('[]', []),
        ('{"abc": "def"}', {"abc": "def"}),
        ('{"value": 42}', {"value": 42}),
        ('{"value": -12.3}', {"value": -12.3}),
        ('{"value": "\\\"Hello\\\" \\n\\b\\f\\r\\t\\u0123"}',
         {"value": '"Hello" \n\b\f\r\t\u0123'}),
        ('["foo", "bar"]', ["foo", "bar"]),
        ('[1, 2, 3]', [1, 2, 3]),
        ('{"value1": true, "value2": false, "value3": null}',
         {"value1": True, "value2": False, "value3": None}),
        ('{"foo": [1, 2, {"bar": 3}]}', {"foo": [1, 2, {"bar": 3}]}),
        ('{"results":[{"gender":"male","name":{"title":"Mr","first":"سینا","last":"موسوی"},"location":{"street":{"number":8134,"name":"میدان امام حسین"},"city":"ارومیه","state":"خوزستان","country":"Iran","postcode":24340,"coordinates":{"latitude":"27.3083","longitude":"-104.2564"},"timezone":{"offset":"0:00","description":"Western Europe Time, London, Lisbon, Casablanca"}},"email":"syn.mwswy@example.com","login":{"uuid":"8a6da152-019a-40b4-80b0-bfafd5281fd7","username":"sadbear764","password":"1947","salt":"ddKNbUrc","md5":"7ff0c750f9b8d7690d50385754a7fe25","sha1":"e140544e222f27a2d0aa809ccb00a1d1ca1fda60","sha256":"fd5dbd61da24a82f48971a6d027400a2fd5b0808fd108764f17d87aaa61774d9"},"dob":{"date":"1996-12-06T21:55:10.574Z","age":24},"registered":{"date":"2013-04-07T05:56:17.049Z","age":7},"phone":"083-15098477","cell":"0998-569-1505","id":{"name":"","value":null},"picture":{"large":"https://randomuser.me/api/portraits/men/94.jpg","medium":"https://randomuser.me/api/portraits/med/men/94.jpg","thumbnail":"https://randomuser.me/api/portraits/thumb/men/94.jpg"},"nat":"IR"}],"info":{"seed":"db5d8d673b395e5a","results":1,"page":1,"version":"1.3"}}',
         {'results': [{'gender': 'male', 'name': {'title': 'Mr', 'first': 'سینا', 'last': 'موسوی'}, 'location': {'street': {'number': 8134, 'name': 'میدان امام حسین'}, 'city': 'ارومیه', 'state': 'خوزستان', 'country': 'Iran', 'postcode': 24340, 'coordinates': {'latitude': '27.3083', 'longitude': '-104.2564'}, 'timezone': {'offset': '0:00', 'description': 'Western Europe Time, London, Lisbon, Casablanca'}}, 'email': 'syn.mwswy@example.com', 'login': {'uuid': '8a6da152-019a-40b4-80b0-bfafd5281fd7', 'username': 'sadbear764', 'password': '1947', 'salt': 'ddKNbUrc', 'md5': '7ff0c750f9b8d7690d50385754a7fe25', 'sha1': 'e140544e222f27a2d0aa809ccb00a1d1ca1fda60', 'sha256': 'fd5dbd61da24a82f48971a6d027400a2fd5b0808fd108764f17d87aaa61774d9'}, 'dob': {'date': '1996-12-06T21:55:10.574Z', 'age': 24}, 'registered': {'date': '2013-04-07T05:56:17.049Z', 'age': 7}, 'phone': '083-15098477', 'cell': '0998-569-1505', 'id': {'name': '', 'value': None}, 'picture': {'large': 'https://randomuser.me/api/portraits/men/94.jpg', 'medium': 'https://randomuser.me/api/portraits/med/men/94.jpg', 'thumbnail': 'https://randomuser.me/api/portraits/thumb/men/94.jpg'}, 'nat': 'IR'}], 'info': {'seed': 'db5d8d673b395e5a', 'results': 1, 'page': 1, 'version': '1.3'}}),
        ("""
        {
            "_id": "5fe977c6eae5f05b1903ee48",
            "index": 0,
            "guid": "f542c7d5-35f5-432d-9ba4-581979a77765",
            "isActive": true,
            "balance": "$2,018.45",
            "picture": "http://placehold.it/32x32",
            "age": 34,
            "eyeColor": "blue",
            "name": "King Hampton",
            "gender": "male",
            "company": "BOILICON",
            "email": "kinghampton@boilicon.com",
            "phone": "+1 (873) 467-3013",
            "address": "971 Calyer Street, Wollochet, New Jersey, 3753",
            "about": "Aute velit consequat tempor exercitation id occaecat sit pariatur irure laboris pariatur veniam fugiat aute. Laborum ea do occaecat culpa labore. Laboris anim commodo ipsum duis eu nulla incididunt in Lorem consectetur cillum in est. Velit amet eiusmod ullamco ut amet. Consectetur elit labore aliqua excepteur. Elit est eu voluptate reprehenderit qui aliqua exercitation ad esse consectetur. Proident quis ex quis proident ad dolor ex sint elit consectetur eu Lorem do proident.\r\n",
            "registered": "2017-01-04T11:20:50 -06:-30",
            "latitude": -66.260402,
            "longitude": 52.599917,
            "tags": [
                "ea",
                "anim",
                "Lorem",
                "veniam",
                "veniam",
                "enim",
                "officia"
            ],
            "friends": [
            {
                "id": 0,
                "name": "Leola Callahan"
            },
            {
                "id": 1,
                "name": "Doyle Ashley"
            },
            {
                "id": 2,
                "name": "Tamara Singleton"
            }
            ],
            "greeting": "Hello, King Hampton! You have 5 unread messages.",
            "favoriteFruit": "strawberry"
        }
  """, {
            "_id": "5fe977c6eae5f05b1903ee48",
            "index": 0,
            "guid": "f542c7d5-35f5-432d-9ba4-581979a77765",
            "isActive": True,
            "balance": "$2,018.45",
            "picture": "http://placehold.it/32x32",
            "age": 34,
            "eyeColor": "blue",
            "name": "King Hampton",
            "gender": "male",
            "company": "BOILICON",
            "email": "kinghampton@boilicon.com",
            "phone": "+1 (873) 467-3013",
            "address": "971 Calyer Street, Wollochet, New Jersey, 3753",
            "about": "Aute velit consequat tempor exercitation id occaecat sit pariatur irure laboris pariatur veniam fugiat aute. Laborum ea do occaecat culpa labore. Laboris anim commodo ipsum duis eu nulla incididunt in Lorem consectetur cillum in est. Velit amet eiusmod ullamco ut amet. Consectetur elit labore aliqua excepteur. Elit est eu voluptate reprehenderit qui aliqua exercitation ad esse consectetur. Proident quis ex quis proident ad dolor ex sint elit consectetur eu Lorem do proident.\r\n",
            "registered": "2017-01-04T11:20:50 -06:-30",
            "latitude": -66.260402,
            "longitude": 52.599917,
            "tags": [
                "ea",
                "anim",
                "Lorem",
                "veniam",
                "veniam",
                "enim",
                "officia"
            ],
            "friends": [
                {
                    "id": 0,
                    "name": "Leola Callahan"
                },
                {
                    "id": 1,
                    "name": "Doyle Ashley"
                },
                {
                    "id": 2,
                    "name": "Tamara Singleton"
                }
            ],
            "greeting": "Hello, King Hampton! You have 5 unread messages.",
            "favoriteFruit": "strawberry"
        })
    )
)
def test_parser(json_string: str, expected: Dict[str, object]) -> None:
    """JSON parser tests"""
    assert json_parser.parse(json_string) == expected
